#define CL_SILENCE_DEPRECATION
#define IN_OPENCL

#include "OpenCL.hxx"
#include "timerCL.hxx"
#include <fstream>
#include <sstream>
#include <iostream>
#include <string>

OpenCL::OpenCL() 
{
  T_Alloc.start();
  platform_id = NULL;
  device_id = NULL;
  
  cl_int errcode;
  cl_uint uret;
  
  errcode = clGetPlatformIDs(1,
			     &platform_id,
			     &uret);
  CheckOpenCL("clGetPlatformIDs");
   
  errcode = clGetDeviceIDs(  platform_id,
			     CL_DEVICE_TYPE_DEFAULT,
			     1, 
			     &device_id,
			     &uret);
  CheckOpenCL("clGetDeviceIDs");

  context = clCreateContext
    (NULL, 1, &device_id, NULL, NULL, &errcode);
  CheckOpenCL("clCreateContext");

#if CL_VERSION_2_0 == 1
  command_queue = clCreateCommandQueueWithProperties
    (context, device_id, NULL, &errcode);
#else
  command_queue = clCreateCommandQueue
    (context, device_id, CL_QUEUE_PROFILING_ENABLE, &errcode);
#endif

  CheckOpenCL("clCreateCommandQueueWithProperties");

  T_Alloc.setQueue(command_queue);
  T_Compile.setQueue(command_queue);
  T_Copy.setQueue(command_queue);
  T_Free.setQueue(command_queue);

  T_Alloc.stop();
}

OpenCL::~OpenCL()
{
  T_Free.start();
  clFlush(command_queue);
  clFinish(command_queue);
  clReleaseCommandQueue(command_queue);
  clReleaseContext(context);
  T_Free.setQueue(NULL);
  T_Free.stop();
}


cl_kernel OpenCL::new_kernel(const char *kernelName,
                             const char *fileName,
                             const char * header)
{
  T_Compile.start();

  cl_int errcode;
  std::string kernel_source = "";

  if (header)
    kernel_source = header;

  std::string fullName = INSTALL_PREFIX;
  fullName += "/";
  fullName += fileName;
  std::ifstream t(fullName.c_str());
  std::stringstream buffer;
  buffer << t.rdbuf();

  kernel_source += buffer.str();
  t.close();

  const char * kernel_src = kernel_source.c_str();
  size_t kernel_len = kernel_source.size();

  cl_program program = clCreateProgramWithSource
    (context, 1, &kernel_src, &kernel_len, &errcode);
  CheckOpenCL("clCreateProgramWithSource");
  
  errcode = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
  if (errcode < 0) {
    size_t len_log = 4999, ret_log;
    char log[len_log+1];
    clGetProgramBuildInfo (program,
                           device_id,
                           CL_PROGRAM_BUILD_LOG,
                           len_log,
                           log,
                           &ret_log);
    std::cerr << log << std::endl;
  }
  CheckOpenCL("clBuildProgram");
  
  // Create the OpenCL kernel
  cl_kernel k =  clCreateKernel(program, kernelName, &errcode);
  CheckOpenCL("clCreateKernel");

  T_Compile.stop();
  return k;
  
}

void OpenCL::free_kernel(cl_kernel & k) {
  T_Free.start();
  clReleaseKernel(k);
  k = NULL;
  T_Free.stop();
}

void OpenCL::info()
{
  size_t param_value_size = 1023;
  char param_value[1024];
  size_t param_value_size_ret;
  cl_platform_info param_name;
  
  param_name = CL_PLATFORM_VERSION;
  clGetPlatformInfo(platform_id,
			  param_name,
			  param_value_size,
			  param_value,
			  &param_value_size_ret);
  std::cerr << "\n\tOpenCL version : " << param_value << std::endl;

  param_name = CL_PLATFORM_NAME;
  clGetPlatformInfo(platform_id,
			  param_name,
			  param_value_size,
			  param_value,
			  &param_value_size_ret);
  std::cerr << "\tPlatform name  : " << param_value << std::endl;
}

cl_mem OpenCL::allocate(int n)
{
  T_Alloc.start();
  cl_mem p;
  size_t bytes = n*sizeof(double);
  cl_int errcode;
  p = clCreateBuffer (context, CL_MEM_READ_WRITE, bytes, NULL, &errcode);
  CheckOpenCL("clCreateBuffer");
  T_Alloc.stop();

  return p;
}

void OpenCL::deallocate(cl_mem &p)
{
  T_Free.start();
  clReleaseMemObject(p);
  T_Free.stop();
}

void OpenCL::memcpyHostToDevice(void *h, cl_mem d, size_t n)
{
  cl_event event;
  T_Copy.start();
  cl_int errcode = clEnqueueWriteBuffer(command_queue, d, CL_TRUE, 0, 
                                          n, h, 0, NULL, &event);
  CheckOpenCL("clEnqueueReadBuffer");
  clWaitForEvents(1, &event); 
  T_Copy.stop();
}

void OpenCL::memcpyDeviceToHost(cl_mem d, void *h, size_t n)
{
  cl_event event;
  T_Copy.start();
  cl_int errcode = clEnqueueReadBuffer(command_queue, d, CL_TRUE, 0, 
                                         n, h, 0, NULL, &event);
  CheckOpenCL("clEnqueueReadBuffer");
  clWaitForEvents(1, &event);
  T_Copy.stop();
}
