/**
* Copyright (C) 2019-2021 Xilinx, Inc
*
* Licensed under the Apache License, Version 2.0 (the "License"). You may
* not use this file except in compliance with the License. A copy of the
* License is located at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
* WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
* License for the specific language governing permissions and limitations
* under the License.
*/

#include <array>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <utility>
#include <vector>

#define CL_TARGET_OPENCL_VERSION 120

#include <CL/cl.h>
#define OCL_CHECK(call)   \
{                         \
    cl_int error = call;  \
    if (error != CL_SUCCESS) {                                                                   \
        printf("%s:%d Error calling " #call ", error code is: %d\n", __FILE__, __LINE__, error); \
        exit(EXIT_FAILURE);                                                                      \
    } \
}

using std::array;
using std::pair;
using std::string;
using std::strstr;
using std::strtok;
using std::vector;

// Wrap any OpenCL API calls that return error code(cl_int)
// with the below macro to quickly check for an error

#define ENUM_CASE(ENUM)     \
    case ENUM:              \
        printf(#ENUM "\n"); \
        break

#define BITFIELD_SETUP(type)                  \
    auto value = convert<type>(field.data()); \
    printf("[ ")

#define BITFIELD_PRINT(FIELD)     \
    if (value & FIELD) {          \
        printf(#FIELD " "); \
    }

#define BITFIELD_END() printf("]\n")

pair<int, const char*> platform_info[] = {{CL_PLATFORM_PROFILE, "profile"},
                                          {CL_PLATFORM_VERSION, "version"},
                                          {CL_PLATFORM_NAME, "name"},
                                          {CL_PLATFORM_VENDOR, "vendor"}};

pair<int, const char*> device_info[] = {
    {CL_DEVICE_TYPE, "type"},
    {CL_DEVICE_NAME, "name"},
    {CL_DEVICE_VENDOR, "vendor"},
    {CL_DRIVER_VERSION, "version"},
    {CL_DEVICE_VERSION, "version"},
    {CL_DEVICE_MAX_COMPUTE_UNITS, "max compute units"},
    {CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, "max work item dimensions"},
    {CL_DEVICE_MAX_WORK_GROUP_SIZE, "max work group size"},
    {CL_DEVICE_MAX_WORK_ITEM_SIZES, "max work item sizes"},
    {CL_DEVICE_MAX_CLOCK_FREQUENCY, "max clock frequency"},
    {CL_DEVICE_ADDRESS_BITS, "address bits"},
    {CL_DEVICE_MAX_MEM_ALLOC_SIZE, "max mem alloc size"},
    {CL_DEVICE_MAX_PARAMETER_SIZE, "max parameter size"},
    {CL_DEVICE_MAX_SAMPLERS, "max samplers"},
    {CL_DEVICE_MEM_BASE_ADDR_ALIGN, "mem base addr align"},
    {CL_DEVICE_GLOBAL_MEM_CACHE_TYPE, "global mem cache type"},
    {CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE, "global mem cacheline size"},
    {CL_DEVICE_GLOBAL_MEM_CACHE_SIZE, "global mem cache size"},
    {CL_DEVICE_GLOBAL_MEM_SIZE, "global mem size"},
    {CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE, "max constant buffer size"},
    {CL_DEVICE_MAX_CONSTANT_ARGS, "max constant args"},
    {CL_DEVICE_LOCAL_MEM_SIZE, "local mem size"},
    {CL_DEVICE_PRINTF_BUFFER_SIZE, "printf buffer size"}};

template <typename T, size_t N>
int sizeof_array(T (&)[N]) {
    return N;
}

constexpr size_t lmax = 1024;
std::string field(lmax+1, '\0');

void print_platform_info(cl_platform_id platform) {
    size_t length;
    void * data = &(field[0]);
    for (int i = 0; i < sizeof_array(platform_info); i++) {
        OCL_CHECK(clGetPlatformInfo(platform, platform_info[i].first, 1023, data, &length));
        printf("platform %-11s: %s\n", platform_info[i].second, field.c_str());
    }
    printf("\n");
}

template <typename T>
T convert(const char* data) {
    return *reinterpret_cast<const T*>(data);
}

void print_device_info(cl_device_id device) {
    size_t length;
    void * data = &(field[0]);
    for (int i = 0; i < sizeof_array(device_info); i++) {
        OCL_CHECK(clGetDeviceInfo(device, device_info[i].first, 1023, data, &length))

        printf("  device %-32s: ", device_info[i].second);
        switch (device_info[i].first) {
            case CL_DEVICE_ADDRESS_BITS:
            case CL_DEVICE_GLOBAL_MEM_CACHELINE_SIZE:
            case CL_DEVICE_MAX_CLOCK_FREQUENCY:
            case CL_DEVICE_MAX_COMPUTE_UNITS:
            case CL_DEVICE_MAX_CONSTANT_ARGS:
            case CL_DEVICE_MAX_READ_IMAGE_ARGS:
            case CL_DEVICE_MAX_SAMPLERS:
            case CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS:
            case CL_DEVICE_MAX_WRITE_IMAGE_ARGS:
            case CL_DEVICE_MEM_BASE_ADDR_ALIGN:
            case CL_DEVICE_MIN_DATA_TYPE_ALIGN_SIZE:
            case CL_DEVICE_PREFERRED_VECTOR_WIDTH_CHAR:
            case CL_DEVICE_PREFERRED_VECTOR_WIDTH_SHORT:
            case CL_DEVICE_PREFERRED_VECTOR_WIDTH_INT:
            case CL_DEVICE_PREFERRED_VECTOR_WIDTH_LONG:
            case CL_DEVICE_PREFERRED_VECTOR_WIDTH_FLOAT:
            case CL_DEVICE_PREFERRED_VECTOR_WIDTH_DOUBLE:
            case CL_DEVICE_PREFERRED_VECTOR_WIDTH_HALF:
            case CL_DEVICE_NATIVE_VECTOR_WIDTH_CHAR:
            case CL_DEVICE_NATIVE_VECTOR_WIDTH_SHORT:
            case CL_DEVICE_NATIVE_VECTOR_WIDTH_INT:
            case CL_DEVICE_NATIVE_VECTOR_WIDTH_LONG:
            case CL_DEVICE_NATIVE_VECTOR_WIDTH_FLOAT:
            case CL_DEVICE_NATIVE_VECTOR_WIDTH_DOUBLE:
            case CL_DEVICE_NATIVE_VECTOR_WIDTH_HALF:
            case CL_DEVICE_PARTITION_MAX_SUB_DEVICES:
            case CL_DEVICE_REFERENCE_COUNT:
                printf("%d\n", convert<cl_uint>(field.data()));
                break;

            case CL_DEVICE_AVAILABLE:
            case CL_DEVICE_COMPILER_AVAILABLE:
            case CL_DEVICE_ENDIAN_LITTLE:
            case CL_DEVICE_ERROR_CORRECTION_SUPPORT:
            case CL_DEVICE_IMAGE_SUPPORT:
            case CL_DEVICE_HOST_UNIFIED_MEMORY:
            case CL_DEVICE_LINKER_AVAILABLE:
            case CL_DEVICE_PREFERRED_INTEROP_USER_SYNC:
                printf("%s\n", convert<cl_bool>(field.data()) ? "true" : "false");
                break;

            case CL_DEVICE_GLOBAL_MEM_CACHE_SIZE:
            case CL_DEVICE_GLOBAL_MEM_SIZE:
            case CL_DEVICE_LOCAL_MEM_SIZE:
            case CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE:
            case CL_DEVICE_MAX_MEM_ALLOC_SIZE:
                printf("%lu\n", convert<cl_ulong>(field.data()));
                break;

            case CL_DEVICE_IMAGE2D_MAX_HEIGHT:
            case CL_DEVICE_IMAGE2D_MAX_WIDTH:
            case CL_DEVICE_IMAGE3D_MAX_DEPTH:
            case CL_DEVICE_IMAGE3D_MAX_HEIGHT:
            case CL_DEVICE_IMAGE3D_MAX_WIDTH:
            case CL_DEVICE_MAX_PARAMETER_SIZE:
            case CL_DEVICE_MAX_WORK_GROUP_SIZE:
            case CL_DEVICE_PROFILING_TIMER_RESOLUTION:
            case CL_DEVICE_PRINTF_BUFFER_SIZE:
            case CL_DEVICE_IMAGE_MAX_BUFFER_SIZE:
            case CL_DEVICE_IMAGE_MAX_ARRAY_SIZE:
                printf("%zu\n", convert<size_t>(field.data()));
                break;

            case CL_DEVICE_MAX_WORK_ITEM_SIZES: {
                
                size_t max_dim = 0;
                OCL_CHECK(clGetDeviceInfo(device, CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, sizeof(max_dim), &max_dim, NULL));
                printf("[ ");
                for (size_t i = 0; i < max_dim; i++) {
                    printf("%ld ", convert<size_t>(&field.data()[i * sizeof(size_t)]));
                }
                printf("]\n");
                
            } break;

            case CL_DEVICE_NAME:
            case CL_DEVICE_PROFILE:
            case CL_DEVICE_VENDOR:
            case CL_DEVICE_VERSION:
            case CL_DRIVER_VERSION:
            case CL_DEVICE_BUILT_IN_KERNELS:
            case CL_DEVICE_OPENCL_C_VERSION:
                printf("%s\n", field.data());
                break;

            case CL_DEVICE_TYPE:
                switch (convert<cl_device_type>(field.data())) {
                    ENUM_CASE(CL_DEVICE_TYPE_ACCELERATOR);
                    ENUM_CASE(CL_DEVICE_TYPE_CPU);
                    ENUM_CASE(CL_DEVICE_TYPE_GPU);
                    ENUM_CASE(CL_DEVICE_TYPE_DEFAULT);
                    default:
                        printf("UNKNOWN\n");
                }
                break;
            case CL_DEVICE_GLOBAL_MEM_CACHE_TYPE:
                switch (convert<cl_device_mem_cache_type>(field.data())) {
                    ENUM_CASE(CL_NONE);
                    ENUM_CASE(CL_READ_ONLY_CACHE);
                    ENUM_CASE(CL_READ_WRITE_CACHE);
                    default:
                        printf("UNKNOWN\n");
                }
                break;
            case CL_DEVICE_LOCAL_MEM_TYPE:
                switch (convert<cl_device_mem_cache_type>(field.data())) {
                    ENUM_CASE(CL_LOCAL);
                    ENUM_CASE(CL_GLOBAL);
                    default:
                        printf("UNKNOWN\n");
                }
                break;
            case CL_DEVICE_PARTITION_PROPERTIES:
                switch (convert<cl_device_mem_cache_type>(field.data())) {
                    case 0:
                        printf("NONE\n");
                        break;
                        ENUM_CASE(CL_DEVICE_PARTITION_EQUALLY);
                        ENUM_CASE(CL_DEVICE_PARTITION_BY_COUNTS);
                        ENUM_CASE(CL_DEVICE_PARTITION_BY_AFFINITY_DOMAIN);
                    default:
                        printf("UNKNOWN\n");
                }
                break;
            default:
                printf("N/A \n");
                continue;
        }
        field[0] = '\0';
    }
}

// This example prints devices available on this machine and their
// corresponding capabilities.
int main(int argc, char** argv) {
    // The following call retrieves the total number of platforms available
    cl_uint platform_count;
  
    clGetPlatformIDs(0, NULL, &platform_count);
    std::vector<cl_platform_id> platforms(platform_count);
    clGetPlatformIDs(platform_count, &(platforms[0]), NULL);

    for (int p = 0; p < (int)platform_count; ++p) {
        print_platform_info(platforms[p]);

        cl_uint device_count = 0;
        clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, 0, NULL, &device_count);
        std::vector<cl_device_id> devices(device_count);
        clGetDeviceIDs(platforms[p], CL_DEVICE_TYPE_ALL, device_count, &(devices[0]), NULL);

        for (int d = 0; d < (int)device_count; ++d) {
            printf("Device %d:\n\n", d);
            print_device_info(devices[d]);
        }
    }
    printf("TEST PASSED\n");

    return EXIT_SUCCESS;
}
